"""
This file is part of eduworld package.

This is a main application window class with Canvas and board and etc

=== LICENSE INFO ===

Copyright (c) 2024 - Stanislav Grinkov

The eduworld package is free software: you can redistribute it
and/or modify it under the terms of the GNU General Public License
as published by the Free Software Foundation, either version 3
of the License, or (at your option) any later version.

The package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with the algoworld package.
If not, see `<https://www.gnu.org/licenses/>`_.
"""

import tkinter as tk
import time
import sys
from concurrent.futures import ThreadPoolExecutor

from .canvas import Canvas


class Application:
    """Defines an Tkinter application with canvas, board and other stuff"""

    def __init__(self, title=""):
        background = "gray75"
        window_width = 800
        window_height = 600
        self.initialized = False
        self.root: tk.Tk = tk.Tk()
        self.canvas: Canvas = Canvas(self.root)
        self.root.title("AlgoWorld" if title == "" else f"AlgoWorld-{title}")
        self.root.configure(background=background)
        self.root.rowconfigure(0, weight=1)
        self.root.columnconfigure(0, weight=1)
        self.root.geometry(f"{window_width}x{window_height}")
        self.root.bind("<Q>", lambda _: self._quit())
        self.root.bind("<q>", lambda _: self._quit())

    def set_board(self, board):
        """Set canvas's board"""
        self.canvas.board = board
        self.initialized = board.initialized

    def _quit(self):
        self.root.quit()
        self.root.quit()

    def run(self) -> None:
        """Runs tkinter TK main loop in a separate thread, to allow
        this main thread to continue the work.
        You have ot call app.shutdown() function to keep the window open
        when your code done its work"""
        if not self.initialized:
            print("Board is not initialized! Exiting!")
            sys.exit(1)
        with ThreadPoolExecutor(max_workers=1) as ex:
            ex.submit(self.root.mainloop)
        self.canvas.redraw()
        self.root.update()
        time.sleep(1)

    def shutdown(self, keep_window: bool = False) -> None:
        """Join tkinter Tk mainloop to the main thread to keep the window open"""
        if keep_window:
            self.root.mainloop()
        else:
            time.sleep(3)



# загрузить команды исполнителя
from eduworld.simple import setup, shutdown, up, down, left, right, put

# подготовить робота
setup(world="lessons/01-move-me")
# выполнить вниз
down()
# выполнить направо
right()
# выполнить направо
right()
# выполнить оставить маяк
put()
# выполнить налево
left()
# выполнить вверх
up()
# выполнить налево
left()
# завершить работу
shutdown(True)



r"""
This file is part of eduworld package.

This class defines Board that can read file based world-boards

=== LICENSE INFO ===

Copyright (c) 2024 - Stanislav Grinkov

The eduworld package is free software: you can redistribute it
and/or modify it under the terms of the GNU General Public License
as published by the Free Software Foundation, either version 3
of the License, or (at your option) any later version.

The package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with the algoworld package.
If not, see `<https://www.gnu.org/licenses/>`_.
"""

import sys
import re
from pathlib import Path

from .tile import Tile
from .board import Board
from .robot import Robot

# pylint: disable=too-few-public-methods


class MapDef:
    """Class encapsulating logic for parsing world file definitions
    World file format defined as follows.
    - It should specify one definition per line in the format
      DEF_NAME: PARAMETERS
    - Empty lines or lines starts with # will be ignored
    - PARAMETERS:
        - Format: name=value. There is no spaces around the equals sign!
        - parameters are separated by space
        - case insensitive
        - can appear in any order
        - unknown parameter names are ignored
    - The accepted DEF_NAME: PARAMETERS pairs
        - size: rows=int cols=int
          both "rows" and "cols" parameters are *required*
        - tile: x=N y=N beepers=N,* temperature=N radiation=N color="" walls=""
            - Required: "x" and "y" parameters
            - Other parameters are optional
            - Parameter "beepers" accepts either special '*' char value
              which means 'inifinite' amount, or int value - amount of beepers
            - Parameter "color" accepts any color value described in Tk docs
              but without spaces: https://tcl.tk/man/tcl8.6/TkCmd/colors.htm
            - Parameter "walls" describe walls around the cell
              t - top, b - bottom, l - left, r - right, f - full
              It accepts only 't', 'b', 'r', 'l', or 'f' chars
              or any combination of them in any order (e.g. 'tbr' or "tlbr")
              Unknown characters are ignored. e.g. "tool" is the same as "tl"
              Duplicate characters are ignored.
    """

    Invalid = 0
    Size = 1
    Tile = 2
    Robot = 3

    def __init__(self, line: str, line_index: int):
        self.type = 0
        self.data = {}
        self._line = line.strip().lower()
        self._index = line_index
        self._fns = self._init_fns()
        self._parse()

    def _init_fns(self):
        def to_int(v):
            return int(v)

        def as_is(v):
            return v

        def to_beepers(v):
            """value of * or -1 means that tile will have inifinite beepers"""
            return -1 if v == "*" else int(v)

        return {"as_is": as_is, "to_int": to_int, "to_beepers": to_beepers}

    def _parse(
        self,
    ) -> None:
        line = self._line
        if not line or line.startswith("#"):
            return
        if line.startswith("map:"):
            self._parse_def("map", MapDef.Size)
            self._validate_size_def()
            return
        if line.startswith("tile:"):
            self._parse_def("tile", MapDef.Tile)
            self._validate_tile_def()
            return
        if line.startswith("robot:"):
            self._parse_def("robot", MapDef.Robot)
            self._validate_robot_def()

    def _validate_robot_def(self):
        to_int = self._fns["to_int"]
        to_beepers = self._fns["to_beepers"]
        pairs = {
            "x": {"fn": to_int, "req": True},
            "y": {"fn": to_int, "req": True},
            "beepers": {"fn": to_beepers, "req": False},
        }
        self._validate_req_data(pairs)

    def _validate_tile_def(self):
        to_int = self._fns["to_int"]
        to_beepers = self._fns["to_beepers"]
        as_is = self._fns["as_is"]
        pairs = {
            "x": {"fn": to_int, "req": True},
            "y": {"fn": to_int, "req": True},
            "radiation": {"fn": to_int, "req": False},
            "temperature": {"fn": to_int, "req": False},
            "beepers": {"fn": to_beepers, "req": False},
            "color": {"fn": as_is, "req": False},
            "walls": {"fn": as_is, "req": False},
        }
        self._validate_req_data(pairs)

    def _validate_size_def(self):
        to_int = self._fns["to_int"]
        pairs = {
            "rows": {"fn": to_int, "req": True},
            "cols": {"fn": to_int, "req": True},
        }
        self._validate_req_data(pairs)

    def _parse_def(self, def_key: str, def_type: int) -> None:
        self.type = def_type
        start = len(def_key) + 1  # to take into account :
        parts = set(re.split(" ", self._line[start:].lstrip()))
        self._parse_parts(parts)

    def _parse_parts(self, parts) -> None:
        for p in parts:
            k, v = re.split("=", p)
            self.data[k] = v

    def _validate_req_data(self, pairs):
        for k, v in pairs.items():
            if v["req"] and k not in self.data:
                raise ValueError(
                    f'Definition "{self._line}" as {self._index} '
                    f"does not contains required '{k}' parameter"
                )
            if k in self.data:
                self.data[k] = v["fn"](self.data[k])

        for k, _ in self.data.items():
            if k not in pairs:
                del self.data[k]


class AlgoWorldBoard(Board):
    """Board loaded from the world file in AlgoWorld World format .aww extension"""

    def __init__(self, world: str):
        super().__init__()
        self.world_path = self._find_world_path(world)
        self._load()

    def _find_world_path(self, world_name):
        if not world_name.endswith(".aww"):
            world_name += ".aww"

        package_path = Path(__file__).absolute().parent / "worlds"
        for worlds_dir in [Path("."), Path("worlds"), package_path]:
            if not worlds_dir.is_dir():
                continue
            world_path = worlds_dir / world_name
            if world_path.is_file():
                return world_path

        raise FileNotFoundError(
            "The specified file was not one of provided worlds.\n"
            "Please store custom worlds in a directory named 'worlds'"
        )

    def _load(self):
        with self.world_path.open(encoding="utf-8") as f:
            for i, line in enumerate(f):
                ld = MapDef(line, i)
                if ld.type == MapDef.Size:
                    self.nrows = ld.data["rows"]
                    self.ncols = ld.data["cols"]
                    self.tiles.set_bounds(self.nrows, self.ncols)
                    self.initialized = True
                if ld.type == MapDef.Tile:
                    x = ld.data["x"]
                    y = ld.data["y"]
                    tile = Tile(**ld.data)
                    self.tiles[x, y] = tile
                if ld.type == MapDef.Robot:
                    x = ld.data.get("x", 1)
                    y = ld.data.get("y", 1)
                    b = ld.data.get("beepers", -1)
                    r = Robot()
                    r.setup(name="default", x=x, y=y, beepers=b)
                    self.add_robot(robot=r)

        if not self.initialized:
            print(
                "ERROR: Map definition does not contain required MAP tag "
                "with the size of the board (rows and cols)"
            )
            sys.exit(1)



"""
Class repesenting a map tile

This file is part of eduworld package.

=== LICENSE INFO ===

Copyright (c) 2024 - Stanislav Grinkov

The eduworld package is free software: you can redistribute it
and/or modify it under the terms of the GNU General Public License
as published by the Free Software Foundation, either version 3
of the License, or (at your option) any later version.

The package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with the algoworld package.
If not, see `<https://www.gnu.org/licenses/>`_.
"""

from typing import Tuple, Dict
from .tile import Tile


Coords = Tuple[int, int]
MapType = Dict[Coords, Tile]


class TileMap:
    """Sparse container for a tiles on a grid
    Technically this container is infinite, but
    at the moment is used as a bounded grid
    """

    def __init__(self):
        self.nrows: int = None
        self.ncols: int = None
        self.tiles: MapType = {}

    def __getitem__(self, xy: Coords) -> Tile:
        if not self.in_range(*xy):
            return Tile(*xy, walls="f")
        return self.tiles.get(xy, Tile(*xy))

    def __setitem__(self, xy: Coords, tile: Tile) -> None:
        self.tiles[xy] = tile

    def set_bounds(self, nrows: int, ncols: int) -> None:
        """Set bounds of the world"""
        self.nrows: int = nrows
        self.ncols: int = ncols

    def in_range(self, x: int, y: int) -> bool:
        """Check if coordinates in map[cols,rows] range"""
        return 0 < y <= self.nrows and 0 < x <= self.ncols

    def has_tile_at(self, xy: Coords) -> bool:
        """this is a crazy way to check if sparse map has tile at coords"""
        return not self.tiles.get(xy, None) is None



r"""
This file is part of eduworld package.

This file contains Generic Robot that can move in four cardinal directions
pickup things, and sense dangerous areas (e.g. walls)

Coordinate system for the robot is defined as follows

   1 2 3 4 5
  +- - - - -> x
1 |
  |
  |
  |
5 |
y v

=== LICENSE INFO ===

Copyright (c) 2024 - Stanislav Grinkov

The eduworld package is free software: you can redistribute it
and/or modify it under the terms of the GNU General Public License
as published by the Free Software Foundation, either version 3
of the License, or (at your option) any later version.

The package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with the algoworld package.
If not, see `<https://www.gnu.org/licenses/>`_.
"""

import time

from .board import Board


class RobotError(RuntimeError):
    """Raised on wrong execution of a command - e.g. hitting a wall."""


# pylint: disable=too-many-public-methods


class Robot:
    """Generic Robot that can move in four cardinal directions,
    pickup and place things, and sense walls, paint, and other things"""

    step_delay = 1

    def __init__(self):
        self.name = f"robot-{time.monotonic()}"
        self.board: Board = None
        self.canvas = None
        self.color = "green"
        self.x: int = 1
        self.y: int = 1
        self.beepers: int = -1

    def setup(
        self, x: int, y: int, beepers: int = -1, color: str = "green", name: str = None
    ):
        """Set up robot color, position, and beepers"""
        self.color = color
        self.x: int = x
        self.y: int = y
        self.beepers: int = beepers
        if name is not None:
            self.name = name

    def __repr__(self) -> str:
        b = "*" if self.beepers == -1 else self.beepers
        return f"name: {self.name}; x: {self.x}; y: {self.y}; beepers: {b}"

    def put(self) -> None:
        """Robot puts the beeper"""
        time.sleep(Robot.step_delay)
        if not self.has_beepers():
            raise RobotError("Out of beepers!")
        if self.beepers != -1:
            self.beepers -= 1
        self.board.place_beeper(self.x, self.y)
        self.canvas.redraw()

    def pickup(self) -> None:
        """Robot picks up the beeper (if tile has any)"""
        time.sleep(Robot.step_delay)
        if not self.board.pickup_beeper(self.x, self.y):
            raise RobotError("Tile has no beepers!")
        if self.beepers != -1:
            self.beepers += 1
        self.canvas.redraw()

    def has_beepers(self) -> bool:
        """True if robot has some or unlimited beepers"""
        return self.beepers != 0

    def next_to_beeper(self) -> bool:
        """Return whether or not tile has beepers"""
        return self.board.has_beepers(self.x, self.y)

    def paint(self, color: str) -> None:
        """Paint the tile ora... I mean in any available color"""
        time.sleep(Robot.step_delay)
        self.board.paint_tile(self.x, self.y, color)
        self.canvas.redraw()

    def tile_color(self) -> str:
        """Return color of the tile"""
        return self.board.read_tile_color(self.x, self.y)

    def tile_radiation(self) -> int:
        """Return tile radiation"""
        return self.board.read_tile_radiation(self.x, self.y)

    def tile_temperature(self) -> int:
        """Return tile temperature"""
        return self.board.read_tile_temperature(self.x, self.y)

    def left(self) -> None:
        """Robot moves left"""
        time.sleep(Robot.step_delay)
        if self.left_is_wall():
            raise RobotError("Can't move left. Something is on the way.")
        self.x = self.x - 1
        self.canvas.redraw()

    def right(self) -> None:
        """Robot moves right"""
        time.sleep(Robot.step_delay)
        if self.right_is_wall():
            raise RobotError("Can't move right. Something is on the way.")
        self.x = self.x + 1
        self.canvas.redraw()

    def up(self) -> None:
        """Robot moves up"""
        time.sleep(Robot.step_delay)
        if self.up_is_wall():
            raise RobotError("Can't move up. Something is on the way.")
        self.y = self.y - 1
        self.canvas.redraw()

    def down(self) -> None:
        """Robot moves down"""
        time.sleep(Robot.step_delay)
        if self.down_is_wall():
            raise RobotError("Can't move down. Something is on the way.")
        self.y = self.y + 1
        self.canvas.redraw()

    def up_is_wall(self):
        """Test if tile above the robot is blocked by wall"""
        return self.board.move_up_blocked(self.x, self.y)

    def down_is_wall(self):
        """Test if tile below the robot is blocked by wall"""
        return self.board.move_down_blocked(self.x, self.y)

    def left_is_wall(self):
        """Test if tile left to the robot is blocked by wall"""
        return self.board.move_left_blocked(self.x, self.y)

    def right_is_wall(self):
        """Test if tile right to the robot is blocked by wall"""
        return self.board.move_right_blocked(self.x, self.y)


#    def up_is_free(self):
#        """Test if tile above the robot is free to move"""
#        return not self.up_is_wall()
#
#    def down_is_free(self):
#        """Test if tile below the robot is free to move"""
#        return not self.down_is_wall()
#
#    def left_is_free(self):
#        """Test if tile left to the robot is free to move"""
#        return not self.left_is_wall()
#
#    def right_is_free(self):
#        """Test if tile right to the robot is free to move"""
#        return not self.right_is_wall()



r"""
This package provides the following functions (__all__) for writing oop
programs with EduWorld Robots

This file is part of eduworld package

=== LICENSE INFO ===

Copyright (c) 2024 - Stanislav Grinkov

The eduworld package is free software: you can redistribute it
and/or modify it under the terms of the GNU General Public License
as published by the Free Software Foundation, either version 3
of the License, or (at your option) any later version.

The package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with the algoworld package.
If not, see `<https://www.gnu.org/licenses/>`_.
"""

from .application import Application
from .board import Board
from .algoworldboard import AlgoWorldBoard
from .canvas import Canvas
from .robot import Robot


__all__ = ["Application", "Board", "Canvas", "AlgoWorldBoard", "Robot"]



r"""
This file is part of eduworld package.

=== LICENSE INFO ===

Copyright (c) 2024 - Stanislav Grinkov

The eduworld package is free software: you can redistribute it
and/or modify it under the terms of the GNU General Public License
as published by the Free Software Foundation, either version 3
of the License, or (at your option) any later version.

The package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with the algoworld package.
If not, see `<https://www.gnu.org/licenses/>`_.
"""

import tkinter as tk

from .tile import Tile
from .robot import Robot

PADDING = 20


# pylint: disable=too-many-ancestors, too-many-instance-attributes
class Canvas(tk.Canvas):
    """Class for drawing the board (with all the stuff) and robots"""

    def __init__(self, parent):
        super().__init__(parent, width=320, height=240, background="gray90")
        self.root = parent
        self.cell_size = 0
        self.half_cell = 0
        self.x_left = 0
        self.y_top = 0
        self.x_right = 0
        self.y_bottom = 0
        self.board = None
        self.grid(column=0, row=0, sticky="NEWS")
        self.bind("<Configure>", lambda _: self.redraw())

    def redraw(self) -> None:
        """Redraw the board and the robots on window resize"""
        if self.board is None or not self.board.initialized:
            return
        self._delete_all()
        self._calculate_cell_size()
        self._draw_tiles()
        self._draw_robots()
        self._draw_border()
        self._draw_box_numbers()
        self.root.update()

    def _delete_all(self):
        self.delete("numbers")
        self.delete("wall")
        self.delete("cell")
        self.delete("beeper")

    def _draw_box_numbers(self):
        self._draw_xy_text()
        self._draw_row_numbers()
        self._draw_col_numbers()

    def _draw_xy_text(self):
        qc = self.half_cell * 0.5
        x = self.x_left
        y = self.y_top
        tags = ("numbers",)
        self.create_text(x - qc, y, text=r"y", font="Mono 12 bold", tags=tags)
        self.create_text(x, y - qc, text=r"x", font="Mono 12 bold", tags=tags)

    def _draw_col_numbers(self):
        y = self.y_top - self.half_cell * 0.5
        x_shift = self.half_cell
        tags = ("numbers",)
        for c in range(0, self.board.ncols):
            x = self.x_left + self.cell_size * c + x_shift
            self.create_text(x, y, text=str(c + 1), font="Mono 12 bold", tags=tags)

    def _draw_row_numbers(self):
        x = self.x_left - self.half_cell * 0.5
        y_shift = self.half_cell * 0.95
        tags = ("numbers",)
        for r in range(0, self.board.nrows):
            y = self.y_top + self.cell_size * r + y_shift
            self.create_text(x, y, text=str(r + 1), font="Mono 12 bold", tags=tags)

    def _draw_tiles(self):
        """Draw board tiles"""
        for y in range(1, self.board.nrows + 1):
            dy = (y - 1) * self.cell_size + self.y_top
            for x in range(1, self.board.ncols + 1):
                dx = (x - 1) * self.cell_size + self.x_left
                tile: Tile = self.board.tiles[x, y]
                self._draw_cell(tile, dx, dy)
                self._draw_walls(tile, dx, dy)
                if tile.has_beepers():
                    self._draw_beepers(tile, dx, dy)
                    self._draw_beeper_text(tile, dx, dy)

    def _draw_beepers(self, tile, shift_x, shift_y) -> None:
        r"""Draw beeper outline
        0+shift         x (cell_size) + shift
        +---------------+
        | os+innr_shft  |  os - oval start (x, y)
        |   + _-=-_     |  oe - oval end   (x, y)
        |    /     \    |
        |   |-diam--|   |  diam - oval diameter
        |    \     /    |  innr_shft = cell_size - od
        |     --=-- +   |
        |           oe  |
        +---------------+
        y (cell_size) + shift
        """
        name = f"b-{tile.x}-{tile.y}"
        diam = self.cell_size * 0.85
        ishift = self.cell_size - diam
        os = [shift_x + ishift, shift_y + ishift]
        oe = [shift_x + diam, shift_y + diam]
        c = os + oe
        tags = ("beeper", name)
        self.create_oval(c, width=2, outline="black", fill="gray95", tags=tags)

    def _draw_beeper_text(self, tile, shift_x, shift_y) -> None:
        name = f"b-{tile.x}-{tile.y}-t"
        hc_x = self.half_cell + shift_x
        hc_y = self.half_cell + shift_y
        tags = ("beeper", name)
        count = "*" if tile.beepers == -1 else str(tile.beepers)
        self.create_text(hc_x, hc_y, text=count, font="Mono 12 bold", tags=tags)

    def _draw_walls(self, tile: Tile, shift_x, shift_y) -> None:
        """Draw cell walls.
        mall cross in the center of the cell
        coordinates are relative to the cell_size
          0+shift         x (cell_size) + shift
          +===============+
          H               H
          H               H
          H               H
          H       +       H
          H               H
          H               H
          H               H
          +===============+
          y (cell_size) + shift
        """
        wall_width = 6
        hw = wall_width / 2

        wall_x_len = shift_x + self.cell_size + hw
        wall_y_len = shift_y + self.cell_size + hw
        tag = ("wall",)

        if tile.top_is_wall():
            coords = [shift_x - hw, shift_y, wall_x_len, shift_y]
            self.create_line(coords, width=wall_width, tags=tag)
        if tile.bottom_is_wall():
            bottom_y = shift_y + self.cell_size
            coords = [shift_x - hw, bottom_y, wall_x_len, bottom_y]
            self.create_line(coords, width=wall_width, tags=tag)
        if tile.left_is_wall():
            coords = [shift_x, shift_y - hw, shift_x, wall_y_len]
            self.create_line(coords, width=wall_width, tags=tag)
        if tile.right_is_wall():
            right_x = shift_x + self.cell_size
            coords = [right_x, shift_y - hw, right_x, wall_y_len]
            self.create_line(coords, width=wall_width, tags=tag)

    def _draw_cell(self, tile, shift_x, shift_y):
        r"""Create cell background with small cross in the center of the cell
        coordinates are relative to the cell_size
        shift     hc      x (cell_size) + shift
          +-------|-------+
          |       |       |
          |               |
          |       |  ---------\  line_start
        hc----  --+--     |   ~  line_length
          |       |  ---------/  line_end
          |               |
          |               |
          +---------------+
          y (cell_size) + shift
        """
        tags = ("cell",)
        line_length = 0.15 * self.cell_size
        pad = 0.05 * self.cell_size
        hc_x = self.half_cell + shift_x
        hc_y = self.half_cell + shift_y
        line_start_x = (self.cell_size - line_length) / 2 + shift_x
        line_end_x = line_start_x + line_length
        line_start_y = (self.cell_size - line_length) / 2 + shift_y
        line_end_y = line_start_y + line_length
        cross = [line_start_x, hc_y, line_end_x, hc_y]  # horizontal
        cross += [hc_x, hc_y]  # center
        cross += [hc_x, line_start_y, hc_x, line_end_y]  # vertical
        rect = [
            shift_x + pad,
            shift_y + pad,
            shift_x + self.cell_size - pad,
            shift_y + self.cell_size - pad,
        ]
        self.create_rectangle(rect, width=1, fill=tile.color, outline="", tags=tags)
        self.create_line(cross, width=1, fill="gray25", tags=tags)

    def _draw_border(self) -> None:
        """Draw border around the board to simulate the walled garden"""
        self.delete("border")
        coords = [self.x_left, self.y_top, self.x_right, self.y_bottom]
        self.create_rectangle(coords, width=6, tags=("border",))

    def _calculate_cell_size(self) -> None:
        """Calculate the w/h of the tile in pixels"""
        h_size = (self.winfo_height() - 2 * PADDING) / self.board.nrows
        w_size = (self.winfo_width() - 2 * PADDING) / self.board.ncols
        self.cell_size = min(h_size, w_size, 80)
        self.half_cell = self.cell_size / 2
        bw = self.cell_size * self.board.ncols
        bh = self.cell_size * self.board.nrows
        self.x_left = (self.winfo_width() - bw) / 2
        self.y_top = (self.winfo_height() - bh) / 2
        self.x_right = self.x_left + bw
        self.y_bottom = self.y_top + bh

    def _draw_robots(self):
        """Draw all robots"""
        for r in self.board.robots:
            if r.canvas is None:
                r.canvas = self
            self._draw_robot(r)

    def _draw_robot(self, r: Robot):
        """Draw robot on the canvas"""
        shift_x = (r.x - 1) * self.cell_size + self.x_left
        shift_y = (r.y - 1) * self.cell_size + self.y_top
        fill = r.color
        tag = (r.name,)
        coords = [0.3, 0, 0, 0.3, 0.3, 0.6, 0.6, 0.3]
        coords = [v * self.cell_size for v in coords]

        self.delete(r.name)
        self.create_polygon(coords, width=1, outline="gray25", fill=fill, tags=tag)
        self.moveto(r.name, shift_x, shift_y)
        self.move(r.name, self.cell_size * 0.18, self.cell_size * 0.18)



r"""
This file is part of eduworld package.

This class defines sparse board with tiles

Board coordinate system is defined as follows
   1 2 3 4 5
  +- - - - -> x
1 |
2 |
3 |
4 |
5 |
  v
  y

=== LICENSE INFO ===

Copyright (c) 2024 - Stanislav Grinkov

The eduworld package is free software: you can redistribute it
and/or modify it under the terms of the GNU General Public License
as published by the Free Software Foundation, either version 3
of the License, or (at your option) any later version.

The package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with the algoworld package.
If not, see `<https://www.gnu.org/licenses/>`_.
"""

from enum import Enum, unique

from .tile import Tile
from .tilemap import TileMap


@unique
class Direction(Enum):
    """Class representing robot move direction"""

    UP = (0, -1)
    DOWN = (0, 1)
    LEFT = (-1, 0)
    RIGHT = (1, 0)


class Board:
    """Base class for storing a robot's world with items, walls, and other stuff
    and functions for interacting with it"""

    def __init__(self):
        self.nrows: int = 0
        self.ncols: int = 0
        self.tiles: TileMap = TileMap()
        self.initialized: bool = False
        self.robots = []

    def add_robot(self, robot):
        """Add robot to robots collection for drawing"""
        if robot not in self.robots:
            self.robots.append(robot)
            robot.board = self

    def get_default_robot(self):
        """Return default robot"""
        return self.get_robot("default")

    def get_robot(self, name: str):
        """Return robot by its name, None if not found"""
        for r in self.robots:
            if r.name == name:
                return r
        return None

    def __repr__(self):
        s = ""
        for y in range(1, self.nrows + 1):
            s += "|"
            for x in range(1, self.ncols + 1):
                s += self.tiles[x, y].short_repr()
            s += "|\n"
        return s

    def has_tile_at(self, x: int, y: int):
        """Test if we have tile at coordinates"""
        return self.tiles.has_tile_at((x, y))

    def read_tile_temperature(self, x, y) -> int:
        """Measure tile temperature"""
        return self.tiles[x, y].temperature

    def read_tile_radiation(self, x, y) -> int:
        """Measure tile radiation"""
        return self.tiles[x, y].radiation

    def read_tile_color(self, x, y) -> str:
        """Return color of the tile"""
        return self.tiles[x, y].color

    def paint_tile(self, x, y, color):
        """Paint tile with color"""
        if not self.has_tile_at(x, y):
            self.tiles[x, y] = Tile(x, y)
        self.tiles[x, y].color = color

    def place_beeper(self, x, y):
        """Place one beeper on the tile"""
        if not self.has_tile_at(x, y):
            self.tiles[x, y] = Tile(x, y)
        self.tiles[x, y].add_beeper()

    def has_beepers(self, x, y) -> bool:
        """True if tile has any amount of beepers"""
        return self.tiles[x, y].has_beepers()

    def pickup_beeper(self, x, y) -> bool:
        """Pick up one beeper from the tile. return False if not possible"""
        tile = self.tiles[x, y]
        if not tile.has_beepers():
            return False
        return tile.remove_beeper()

    def move_up_blocked(self, sx, sy, /) -> bool:
        """Test if we can move up from tile[sx, sy]"""
        return self._move_is_blocked(sx, sy, Direction.UP)

    def move_down_blocked(self, sx, sy, /) -> bool:
        """Test if we can move down from tile[sx, sy]"""
        return self._move_is_blocked(sx, sy, Direction.DOWN)

    def move_left_blocked(self, sx, sy, /) -> bool:
        """Test if we can move up from tile[sx, sy]"""
        return self._move_is_blocked(sx, sy, Direction.LEFT)

    def move_right_blocked(self, sx, sy, /) -> bool:
        """Test if we can move up from tile[sx, sy]"""
        return self._move_is_blocked(sx, sy, Direction.RIGHT)

    def _move_is_blocked(self, sx, sy, delta: Direction) -> bool:
        """Use it to test if robot can move from source tile (sx, sy)
        to dest tile. Dest tile is defined using delta by x and y
        """
        dx = sx + delta.value[0]
        dy = sy + delta.value[1]
        t_curr = self.tiles[sx, sy]
        t_next = self.tiles[dx, dy]
        # now let's check the direction
        if delta == Direction.UP:
            return t_curr.top_is_wall() or t_next.bottom_is_wall()
        if delta == Direction.DOWN:
            return t_curr.bottom_is_wall() or t_next.top_is_wall()
        if delta == Direction.LEFT:
            return t_curr.left_is_wall() or t_next.right_is_wall()
        if delta == Direction.RIGHT:
            return t_curr.right_is_wall() or t_next.left_is_wall()
        return True



"""
Class repesenting a map tile

This file is part of eduworld package.

=== LICENSE INFO ===

Copyright (c) 2024 - Stanislav Grinkov

The eduworld package is free software: you can redistribute it
and/or modify it under the terms of the GNU General Public License
as published by the Free Software Foundation, either version 3
of the License, or (at your option) any later version.

The package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with the algoworld package.
If not, see `<https://www.gnu.org/licenses/>`_.
"""

# pylint: disable=too-many-arguments


class Tile:
    """Class representing a map tile (with items, beepers, walls, etc)"""

    FULL_WALL = 15
    TOP_WALL = 8
    RIGHT_WALL = 4
    LEFT_WALL = 2
    BOTTOM_WALL = 1

    def __init__(
        self,
        x: int,
        y: int,
        walls: str = "",
        beepers: int = 0,
        radiation: int = 0,
        temperature: int = 0,
        color: str = "",
    ):
        self.x: int = x
        self.y: int = y
        # using -1 to represent inf count of beepers is not super clear, but ...
        self.beepers: int = beepers
        self.radiation: int = radiation
        self.temperature: int = temperature
        self.color: str = color
        self.walls: int = self._parse_walls(walls)

    def _wall_repr(self) -> str:
        s: str = ""
        if self.top_is_wall():
            s += "t"
        if self.bottom_is_wall():
            s += "b"
        if self.left_is_wall():
            s += "l"
        if self.right_is_wall():
            s += "r"
        return s

    def __repr__(self):
        # tile: x y beepers=n,* temp=n radiation=n color=color walls=tblr
        b = "*" if self.beepers == -1 else self.beepers
        return (
            f"tile: x={self.x} y={self.y} "
            f"beepers={b} temp={self.temperature} "
            f"radiation={self.radiation} "
            f"color={self.color} "
            f"walls={self._wall_repr()}"
        )

    def short_repr(self) -> str:
        """Return short one char representation of the tile"""
        if self.has_walls():
            return "#"
        if self.has_beepers():
            b = self.beepers
            if b == -1:
                return "*"
            if b >= 10:
                return "+"
            return str(b)
        return " "

    def has_beepers(self) -> bool:
        """Test if tile has any beepers on it"""
        return self.beepers == -1 or self.beepers > 0

    def add_beeper(self):
        """Increase count of beepers on the tile (if not inf)"""
        if self.beepers != -1:
            self.beepers += 1

    def remove_beeper(self) -> bool:
        """Decrease count of beepers on the tile (if not inf)"""
        if self.beepers == 0:
            return False
        if self.beepers == -1:
            return True
        self.beepers -= 1
        return True

    def has_walls(self) -> bool:
        """Test if tile has any walls"""
        return self.walls > 0

    def top_is_wall(self) -> bool:
        """Return if top is wall"""
        return self.walls & Tile.TOP_WALL != 0

    def bottom_is_wall(self) -> bool:
        """Return if bottom is wall"""
        return self.walls & Tile.BOTTOM_WALL != 0

    def left_is_wall(self) -> bool:
        """Return if left is wall"""
        return self.walls & Tile.LEFT_WALL != 0

    def right_is_wall(self) -> bool:
        """Return if right is wall"""
        return self.walls & Tile.RIGHT_WALL != 0

    def _parse_walls(self, walls: str) -> int:
        w = 0
        for c in walls.lower():
            if c == "e":
                return 0
            elif c == "f":
                return Tile.FULL_WALL
            elif c == "t":
                w |= Tile.TOP_WALL
            elif c == "b":
                w |= Tile.BOTTOM_WALL
            elif c == "l":
                w |= Tile.LEFT_WALL
            elif c == "r":
                w |= Tile.RIGHT_WALL
        return w



# oop version
from eduworld import Application, Board, AlgoWorldBoard, Robot

Robot.step_delay = 0.0

a: Application = Application()

# this can be substituted by world-gen class instead of file-reader based
b: Board = AlgoWorldBoard(world="demo-world")
a.set_board(board=b)

r: Robot = b.get_default_robot()

a.run()

r.put()
r.right()
r.put()
r.right()
r.paint(color="orange")

if r.tile_color() == "red":
    r.paint(color="green")

if r.tile_color() == "orange":
    r.paint(color="dodgerblue")

r.left()
r.pickup()
r.left()
r.pickup()

if r.down_is_wall():
    r.up()

# todo final pos validation (define it through world file ana validate on exit)
a.shutdown()



r"""
This package provides the following functions defined in __all__ for writing
simple procedural style programs with AlgoWorld Robots

This file is part of eduworld package

=== LICENSE INFO ===

Copyright (c) 2024 - Stanislav Grinkov

The eduworld package is free software: you can redistribute it
and/or modify it under the terms of the GNU General Public License
as published by the Free Software Foundation, either version 3
of the License, or (at your option) any later version.

The package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with the algoworld package.
If not, see `<https://www.gnu.org/licenses/>`_.
"""

import sys
import tkinter as tk
from eduworld import Application, AlgoWorldBoard, Robot

app: Application = Application()
r: Robot = Robot()
initialized: bool = False


def _hide_tk_err(f):
    def w(*args, **kwargs):
        try:
            f(*args, **kwargs)
        except tk.TclError:
            sys.exit(1)

    return w


def step_delay(delay) -> None:
    """Set the delay between robot steps"""
    Robot.step_delay = min(max(0, delay), 2)


def setup(world: str) -> None:
    """Setup the board and the robot"""
    global initialized
    global r
    if not initialized:
        initialized = True
        board = AlgoWorldBoard(world=world)
        app.set_board(board=board)
        r = board.get_default_robot()
        app.run()


def shutdown(keep_window: bool = False) -> None:
    """Shut down the app"""
    app.shutdown(keep_window)


@_hide_tk_err
def up() -> None:
    """Move default robot up"""
    r.up()


@_hide_tk_err
def down() -> None:
    """Move default robot down"""
    r.down()


@_hide_tk_err
def left() -> None:
    """Move default robot left"""
    r.left()


@_hide_tk_err
def right() -> None:
    """Move default robot right"""
    r.right()


@_hide_tk_err
def put() -> None:
    """Put beeper down"""
    r.put()


@_hide_tk_err
def pickup() -> None:
    """Pickup beeper"""
    r.pickup()


@_hide_tk_err
def paint(color: str = "tan2") -> None:
    """Paint tile with color"""
    r.paint(color=color)


def has_beepers() -> bool:
    """Test if robot has beepers"""
    return r.has_beepers()


def next_to_beeper() -> bool:
    """Test if tile the robot on has beepers"""
    return r.next_to_beeper()


def tile_color() -> str:
    """Return tile color"""
    return r.tile_color()


def tile_radiation() -> int:
    """Return tile radiation"""
    return r.tile_radiation()


def tile_temperature() -> int:
    """Return tile temperature"""
    return r.tile_temperature()


def up_is_wall() -> bool:
    """Test if up of robot is wall"""
    return r.up_is_wall()


def down_is_wall() -> bool:
    """Test if down of robot is wall"""
    return r.down_is_wall()


def left_is_wall() -> bool:
    """Test if left of robot is wall"""
    return r.left_is_wall()


def right_is_wall() -> bool:
    """Test if right of robot is wall"""
    return r.right_is_wall()


__all__ = [
    # setup
    "setup",
    "shutdown",
    "step_delay",
    # movement
    "up",
    "down",
    "left",
    "right",
    # check movement
    "up_is_wall",
    "down_is_wall",
    "left_is_wall",
    "right_is_wall",
    # color paint
    "tile_color",
    "paint",
    # beepers
    "put",
    "pickup",
    "has_beepers",
    "next_to_beeper",
    # other sensors
    "tile_radiation",
    "tile_temperature",
]
